#!/usr/bin/env python
"""
Defines the auxiliary functions handling lattices
"""

import numpy as np


def grid_lattice(rows: int, columns: int, distance: float) -> dict:
	"""
	Returns a dictionary corresponding to a grid lattice.
	As the Pulser Devices limit the max distance from the origin, it is preferred to use grid_lattice_centred
	Args:
		- rows: nbr of rows
		- columns: nbr of columns
		- distance: distance between two atoms (nm)
	Output:
		- a dictionary that links the name of the atom with its position
	"""
	q_dict = {}
	for i in range(rows):
		for j in range(columns):
			q_dict["{}".format(i*columns + j)] = np.array([distance*i, distance*j])
	return q_dict


def grid_lattice_centred(dim: int, distance: float) -> dict:
	"""
	Returns a dictionary corresponding to a grid lattice whose center is in 0.
	Args:
			- rows: nbr of rows
			- columns: nbr of columns
			- distance: distance between two atoms (nm)
	Output:
			- a dictionary that links the name of the atom with its position
	"""
	offset = distance * (dim - 1) / 2

	q_dict = {}
	for i in range(dim):
		for j in range(dim):
			q_dict["{}".format(i * dim + j)] = np.array([distance * i, distance * j]) - offset

	return q_dict


def triangle_lattice(edge_size: int, distance: float) -> dict:
	"""
	Returns a dictionary corresponding to a lattice shaped as a triangle
	filled with triangles of same dimension.
	Args:
		- edge_size: nbr of atoms in the largest row
		- distance: distance between two atoms (nm)
	Output:
		- a dictionary that links the name of the atom with its position
	""" 
	q_dict, index = {}, 0
	for i in range(edge_size+1):
		for j in range(i):
			q_dict["{}".format(index)] = np.array([i*distance, j*distance])
			index += 1
	return q_dict


def circle_lattice(nbr_atoms: int, distance: float) -> dict:
	"""
	Returns a dictionary corresponding to a lattice shaped as a circle
	where each point is made of atoms.
	Args:
		- nbr_atoms: distance between two atoms
		- distance: distance between an atom and its nearest neighbours
	Output:
		- a dictionary that links the name of the atom with its position
	"""
	q_dict = {}
	angle_step = 2*np.pi/nbr_atoms
	radius = distance/(np.cos(angle_step)*angle_step)
	for i in range(nbr_atoms):
		q_dict["{}".format(i)] = np.array([radius*np.cos(i*angle_step), radius*np.sin(i*angle_step)])
	return q_dict



#!/usr/bin/env python
"""
Defines the abstract classes Forecaster and Hyper
"""
from naref.modules import metrics, constants, data
from math import prod
from itertools import product
from collections import defaultdict
import csv


class Forecaster:
	"""
	Abstract class containing functions implemented by QRC, CRC, LinReg classes
	"""
	header = []

	def __init__(
			self,
			input_type: str = 'unknown',
			train_len: int = -1,
			test_len: int = -1,
			verbose: bool = constants.VERBOSE):
		self.name = "AFR"  # String identifier of the Forecaster type
		self.longname = "Abstract Forecaster"  # String explicit name of the Forecaster type
		self.y_data_train: list = []  # Contains each predictable point of the train TS (all but the sample_len first)
		self.y_pred_train: list = []  # Filled with the predicted TS after the training. Ideally equal to y_data_train.
		self.y_data_test: list = []  # Contains each predictable point of the test TS (all but the sample_len first)
		self.y_pred_test: list = []  # Filled with the predicted TS after the training. Ideally equal to y_data_test.
		self.rmtrain: dict = {}  # result of metrics for the training phase
		self.rmtest: dict = {}  # result of metrics for the testing phase
		self.input_type: str = input_type  # Clue on the type of data that will be used to build_model
		self.train_len: int = train_len  # Will contain the real_train_length
		self.test_len: int = test_len  # Will contain the real_test_length
		self.verbose: bool = verbose  # verbosity level

	def get_scores(self) -> None:
		"""
		Calculate train and test score of the quantum reservoir and updates rmtest & rmtrain
		"""
		self.rmtrain = metrics.errors(self.y_pred_train, self.y_data_train)
		self.rmtest = metrics.errors(self.y_pred_test, self.y_data_test)

		print(f"{self.name} Train: {self.rmtrain}")
		print(f"{self.name} Test: {self.rmtest}")

	def announce_build(self) -> None:
		"""
		Displays a message stating that we are building the model.
		"""
		if self.verbose:
			print(f"Building {self.longname}...")

	def get_error_train(self) -> float:
		"""
		Evaluates the training error.
		None of the variables used are modified by the testing phase.
		"""
		return constants.ERROR_METRIC(self.y_data_train, self.y_pred_train)

	def get_error_test(self) -> float:
		"""
		Evaluates the testing error.
		"""
		return constants.ERROR_METRIC(self.y_data_test, self.y_pred_test)

	def show_test_prediction(self) -> None:
		"""
		Shows how close are self.y_pred_test and self.y_data_test
		"""
		metrics.show_prediction(self.y_pred_test, self.y_data_test, "Test", self.longname)

	def show_train_prediction(self) -> None:
		"""
		Shows how close are self.y_pred_train and self.y_data_train
		"""
		metrics.show_prediction(self.y_pred_train, self.y_data_train, "Train", self.longname)

	def build_model(self, inp_train: list, inp_test: list) -> list:
		"""
		Updates the inner model and returns the predicted test values after fitting on the train
		Args:
			inp_train: input train datasest (a temporal serie)
			inp_test: input test datasest (a temporal serie)

		Returns:
			the predicted test values after fitting on the train ( y_pred_test )
		"""
		return self.y_pred_test

	def save_data(self, overwrite: bool = False, dic: dict = None, fname: str = 'default', header: list = None):
		"""
			Saves the contents of the qrc to a csv file

			Args:
				overwrite: whether to overwrite the file where we would write data, or just append at the end
				dic: dictionnary containing the contents of a row of data, corresponding to the header
				fname: file name
				header: names of the columns
		"""
		header = header if header is not None else self.header
		dic = dic if dic is not None else {}
		with open(fname, 'w' if overwrite else 'a', encoding='UTF8') as f:
			if overwrite:
				writer = csv.writer(f)
				writer.writerow(header)
			dictwriter_object = csv.DictWriter(f, fieldnames=header)
			dictwriter_object.writerow(dic)
			f.close()


class Hyper:
	"""
	Abstract hyperparameter search class containing functions implemented by HyperC and HyperQ
	"""

	def __init__(
			self,
			sample_len_list=None,
			input_type_list=None,
			train_len_list=None,
			test_len_list=None,
			verbose: bool = constants.VERBOSE):
		self.lists = defaultdict(list, {
			"sample_len": data.sample_len_list if sample_len_list is None else sample_len_list,
			"input_type": data.input_type_list if input_type_list is None else input_type_list,
			"train_len": data.train_len_list if train_len_list is None else train_len_list,
			"test_len": data.test_len_list if test_len_list is None else test_len_list})
		self.k: int = 0
		self.tot_combin: int = 0  # Total number of possible hyperparameters combinations
		self.current: Forecaster = None  # Currently optimised Forecaster
		self.verbose: bool = verbose

	def search(self):
		"""
		Iterates on all possible combinations of parameters in self.lists and adds a line to the relevant csv for each

		Returns:
			all results in a csv file
		"""
		self.k = 0
		self.tot_combin = prod(len(x) for x in self.lists.values())
		print(f"{self.__class__.__name__} search within {self.tot_combin} potential sets of parameters")
		para = product(*self.lists.values())
		for tup in para:
			self.test_parameters(*tup)
		self.print("Hyperparameter search complete")

	def test_parameters(self, *tup):
		"""
		Tests a set of parameters by updating self.current and calling save_parameters
		Args:
			*tup: iterable of parameters

		Returns:
			all results in a csv file
		"""
		raise Exception(f'{self.__class__.__name__} does not implement the test_parameters method.')

	def save_parameters(self):
		"""
		Builds self.current and then appends relevant data in a file
		Returns:
			all results in a csv file
		"""
		self.current.build_model(data.inp_train, data.inp_test)
		self.print(f"{self.k} out of {self.tot_combin}")
		self.current.save_data(self.k < 2)

	def print(self, *a, **k):
		"""
		Wrapper for the print function taking verbosity into account
		"""
		if self.verbose:
			print(*a, **k)



#!/usr/bin/env python
"""
Defines the default data (mackey, sine ...) and initial Forecaster parameters
"""
from reservoirpy.datasets import mackey_glass
from sklearn.preprocessing import MinMaxScaler
import numpy as np


class Dataset:
	# Choose input data: mackey glass ('mackey') or sine curve ('sine')
	input_type = "mackey"

	# MACKEY_GLASS
	# create a simple mackey_glass time series, rescaled between 0.2 and 1 for encoding
	mackey = mackey_glass(2000)
	scaler = MinMaxScaler(feature_range=(0.2, 1))
	scaler.fit(mackey)
	# we noticed that the typo on the following line meant that the rescaling was not applied in our calculations.
	# to make our results reproducible, we decided not to change this feature in the package.
	X = scaler.transform(mackey)

	# SINE
	sine = np.sin(np.linspace(0, 6 * np.pi, 600)).reshape(-1, 1)
	scaler = MinMaxScaler(feature_range=(0.2, 1))
	scaler.fit(sine)
	sine = scaler.transform(sine)

	# split input_type into train and test
	base_data = mackey if input_type == "mackey" else sine
	inp_train = base_data[:250]
	inp_test = base_data[250:280]

	# These are the default fixed hyperparameters
	# Use Comparator.hyper if you want to do an exhaustive hyperparameter search

	train_len = len(inp_train)
	test_len = len(inp_test)
	sample_len = 8

	# Not implemented yet : different values for N_samples, sample_len and reset_rate for train and test.

	# Classical Reservoir Parameters
	nb_neurons = 9

	# Quantum Reservoir Parameters
	# To quickly test that the code works, try 4 atoms.
	# To confirm our results, the hyperparameters are listed in the appendix  of our pdf document
	nb_atoms = 9

	inp_duration = 1000
	N_samples = 1024
	reset_rate = 0.
	geometry = "grid_lattice_centred"
	atom_distance = 15

	# For Hyperparameter search
	# Classical
	nb_neurons_list = [
		4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 25, 30, 35, 40, 50, 60, 70, 80, 90, 100]
	# Quantum
	sample_len_list = [2, 3, 6, 8, 9, 10]
	nb_atoms_list = [9]
	inp_duration_list = [16, 100, 1000]
	N_samples_list = [1024]
	reset_rate_list = [0.]
	geometry_list = ["grid_lattice_centred"]
	atom_distance_list = [4, 8, 11, 13, 14, 15, 20, 35]
	# Potentially usable for both the classical & quantum hyperparameter searches
	# Warning ! It is not recommended to change such parameters for the quantum hyp. search by default,
	# in order not to make a default search too long
	input_type_list = ["sine"]
	train_len_list = [train_len]
	test_len_list = [test_len]

	real_train_len = train_len - sample_len
	real_test_len = test_len - sample_len

	# Energy comparisons
	dataset_sizes = [1000 * i for i in range(1, 10 ** 4 + 1)]
	S = 10
	T_pulse = 1e-6
	N_runs = 52


data = Dataset()



#!/usr/bin/env python
"""
Defines the Comparator class
"""

from naref.modules import data, abstract, metrics, constants
from naref.modules.quantum import QRC
from naref.modules.classical import LinReg, CRC
from collections import defaultdict
from matplotlib import pyplot as plt


class ErrorComparator:

	def __init__(self, active: list = None, auto_train: bool = True):
		"""
		Provides a framework comparing all available Forecasters for all available metrics.
		Args:
			active: list of Forecasters that we want to study
			auto_train: whether to automatically train active Forecasters. Set to false if only interested in energy.
		"""

		self.active = ['CRC', 'LinReg', 'QRC'] if active is None else active  # Forecasters that we want to study
		self.frc = defaultdict(
			abstract.Forecaster,
			{
				'QRC': QRC(
					sample_len=data.sample_len,
					nb_atoms=data.nb_atoms,
					N_samples=data.N_samples,
					inp_duration=data.inp_duration,
					reset_rate=data.reset_rate,
					geometry=data.geometry,
					atom_distance=data.atom_distance,
					input_type=data.input_type,
					test_len=data.real_test_len,
					train_len=data.real_train_len,
					verbose=True),
				'LinReg': LinReg(),
				'CRC': CRC(
					sample_len=data.sample_len,
					nb_neurons=data.nb_neurons)})
		if auto_train:
			self.train()

		# Outer Accessibility
		# Quantum Reservoir
		self.qrc: QRC = self.frc['QRC']
		# Linear Regression
		self.lrg: LinReg = self.frc['LinReg']
		# Classical Reservoir
		self.crc: CRC = self.frc['CRC']

	def train(self) -> None:
		"""
		Builds the Forecaster's attributes according to contents of the data module
		"""
		for forecaster in self.active:
			self.frc[forecaster].build_model(data.inp_train, data.inp_test)

	def compare(self) -> None:
		"""
		Prints errors for all active forecasters and displays their forecast as pyplot curves
		"""
		for forecaster in self.active:
			print("Train error:", self.frc[forecaster].get_error_train())
			self.frc[forecaster].show_train_prediction()
		for forecaster in self.active:
			print("Test error:", self.frc[forecaster].get_error_test())
			self.frc[forecaster].show_test_prediction()


class Comparator(ErrorComparator):

	def __init__(self, dataset_sizes: list = None, *a, **k):
		super().__init__(*a, **k)
		# self.dataset_sizes corresponds to Tilde_L
		self.dataset_sizes = data.dataset_sizes if dataset_sizes is None else dataset_sizes
		self.S = k["S"] if "S" in k.keys() else data.S
		self.T_pulse = k["T_pulse"] if "T_pulse" in k.keys() else data.T_pulse
		self.N_runs = k["N_runs"] if "N_runs" in k.keys() else data.N_runs

		# Default energy lists
		self.E_Fresnel = []
		self.E_Rubi = []
		self.E_GPU = []
		self.E_Joliot = []

	def qtime(self, L: int) -> float:
		"""
		We precisely calculated this value with information given by PASQAL (refresh rate, etc.)
		Args:
			L: dataset size
		Returns:
			Multiplicative coefficient to get the consumption of a quantum device for this dataset size
		"""
		return self.N_runs * 2 ** 5 * (metrics.fct_L(L) - metrics.fct_S(L)) * (1 + metrics.fct_S(L) * self.T_pulse) / 3600

	def rtime(self, L: int) -> float:
		"""
		We could divide by 10 the value of qtime, as it is what is done by default in the Excel file and as this
		computer has a cryostat, but we cannot know for sure. As in the .pdf file, we set it equal to the qtime.
		Args:
			L: dataset size
		Returns:
			Multiplicative coefficient to get the consumption of a quantum device for this dataset size
		"""
		return self.qtime(L)

	def ctime(self, L: int) -> float:
		"""
		Args:
			L: dataset size
		Returns:
			Multiplicative coefficient to get the consumption of a classical GPU device for this dataset size
		"""
		return self.N_runs * (2.496 * 1e5 * L / 10) / metrics.f_GPU * L * (L / 100) / 3600

	def jtime(self, L: int) -> float:
		"""
		Multiplicative coefficient to get the consumption of a typical classical HPC device (Joliot)
		for this dataset size
		Args:
			L: dataset size
		Returns:
			the multiplicative coefficient
		"""
		return self.N_runs * (2.496 * 1e5 * L / 10) / metrics.f_Joliot * L * (L / 100) / 3600

	def energy(self):
		"""
		Computes the energy consumption for each device and each potential dataset size
		And then plots it
		"""
		self.compute_energy()
		self.plot_energy()

	def compute_energy(self):
		"""
		Computes the energy consumption for each device and each potential dataset size
		"""
		print("Computing Fresnel Energy Consumption")
		self.E_Fresnel = [metrics.C_Fresnel * self.qtime(L) for L in self.dataset_sizes]
		print("Computing Rubi Energy Consumption")
		self.E_Rubi = [metrics.C_Rubi * self.rtime(L) for L in self.dataset_sizes]
		print("Computing GPU Energy Consumption")
		self.E_GPU = [metrics.C_GPU * self.ctime(L) for L in self.dataset_sizes]
		print("Computing Joliot Energy Consumption")
		self.E_Joliot = [metrics.C_Joliot * self.jtime(L) for L in self.dataset_sizes]

	def plot_energy(self):
		"""
		Plots the energy that has been computed in compute_energy
		"""
		plt.rcParams.update({'font.size': 22})
		plt.figure(figsize=(15, 6.5))
		plt.title("Evolution of Carbon Emission for Time Series Forecasting \n as Data Size increases")
		plt.xlabel("Size of the dataset")
		plt.ylabel("tCO2 eq")
		plt.xscale("log")
		plt.yscale("log")
		plt.plot(self.dataset_sizes, self.E_Fresnel, label="QRC on Fresnel", color='blue')
		plt.plot(self.dataset_sizes, self.E_Rubi, label="QRC on Rubi", color='yellow')
		plt.plot(self.dataset_sizes, self.E_GPU, label="RNN on GPU", color='green')
		plt.plot(self.dataset_sizes, self.E_Joliot, label="RNN on Joliot-Curie (HPC)", color='red')
		# plt.xticks(self.dataset_sizes)
		plt.legend()
		plt.savefig(f"{constants.MAIN_PATH}/Emission_Comparison_RNN.pdf")
		plt.show()

	def emissions_duration(self):
		"""
		Considering the same values as before, we can derive the different method times and their carbon emissions
		for a typical dataset size
		"""
		L = 350*10**3

		E_Fresnel = metrics.C_Fresnel * self.qtime(L)
		E_Rubi = metrics.C_Rubi * self.rtime(L)
		E_GPU = metrics.C_GPU * self.ctime(L)
		E_Joliot = metrics.C_Joliot * self.jtime(L)

		T_Fresnel = self.qtime(L)
		T_Rubi = self.rtime(L)
		T_GPU = self.ctime(L)
		T_Joliot = self.jtime(L)

		print("Emissions from Fresnel = {} eq tCO2             Duration = {} h".format(E_Fresnel, T_Fresnel))
		print("Emissions from Rubi = {} eq tCO2                Duration = {} h".format(E_Rubi, T_Rubi))
		print("Emissions from GPU = {} eq tCO2 /               Duration = {} h".format(E_GPU, T_GPU))
		print("Emissions from Joliot-Curie HPC = {} eq tCO2     Duration = {} h".format(E_Joliot, T_Joliot))




#!/usr/bin/env python
"""
Defines the CRC and LinearRegression classes (classical TSP)
"""
from sklearn.linear_model import LinearRegression
from naref.modules import abstract, constants
import reservoirpy as rpy

rpy.verbosity(0)  # no need to be too verbose here


def warp_dataset(inp_train: list, inp_test: list) -> (list, list, list, list):
	"""
	Transforms the dataset to fit its own input style in build_baseline_model
	Args:
		inp_train: input train datasest (a temporal serie)
		inp_test: input test datasest (a temporal serie)

	Returns:
		the 4-tuple :
		(x_train = inp_train[:-1]
		y_train = inp_train[1:]
		x_test = inp_test[:-1]
		y_test = inp_test[1:])
	"""
	return inp_train[:-1], inp_train[1:], inp_test[:-1], inp_test[1:]


class LinReg(abstract.Forecaster):

	def __init__(self, verbose: bool = constants.VERBOSE):
		"""
		Class containing Linear Regression model. It is used as the baseline model for comparisons.
		"""
		super().__init__(verbose=verbose)
		self.name: str = "LinReg"
		self.longname: str = "Linear Regression"
		self.model = LinearRegression()  # will contain the last fit to training points

	def build_model(self, inp_train: list, inp_test: list) -> list:
		"""
		Updates the inner model and returns the predicted test values after fitting on the train
		Args:
			inp_train: input train datasest (a temporal serie)
			inp_test: input test datasest (a temporal serie)

		Returns:
			the predicted test values after fitting on the train ( y_pred_test )
		"""
		self.announce_build()
		return self._build_model(*warp_dataset(inp_train, inp_test))

	def _build_model(self, x_train: list, y_train: list, x_test: list, y_test: list) -> list:
		"""
		Updates the inner model and returns the predicted test values after fitting on the train
		Args:
			x_train: training input data
			y_train: training correct output data
			x_test: testing input data
			y_test: testing correct output data

		Returns:
			the predicted test values after fitting on the train ( y_pred_test )
		"""
		self.model.fit(x_train, y_train)
		self.y_data_train: list = y_train
		self.y_pred_train: list = self.model.predict(x_train)
		self.y_data_test: list = y_test
		self.y_pred_test: list = self.model.predict(x_test)
		return self.y_pred_test


class CRC(abstract.Forecaster):
	header = ['sample_len', 'nb_neurons', 'input_type', 'train_len', 'test_len', 'train_mae', 'test_mae']

	def __init__(
			self,
			sample_len: int,
			nb_neurons: int,
			input_type: str = 'unknown',
			train_len: int = -1,
			test_len: int = -1,
			verbose: bool = constants.VERBOSE):
		super().__init__(input_type, train_len, test_len, verbose)
		self.name: str = "CRC"
		self.longname: str = "Classical Reservoir"
		self.sample_len: int = sample_len
		self.nb_neurons: int = nb_neurons
		self.input_type: str = input_type  # Clue on the type of data that will be used to build_model
		self.train_len: int = train_len  # Will contain the real_train_length
		self.test_len: int = test_len  # Will contain the real_test_length

		# reservoir generation
		rpy.set_seed(constants.DEFAULT_SEED)
		reservoir = rpy.nodes.Reservoir(self.nb_neurons, lr=1, sr=0.9)  # Main part of the CRC
		ridge = rpy.nodes.Ridge(ridge=1e-7)  # A single layer of neurons learning with Tikhonov linear regression.
		self.model = reservoir >> ridge

	def build_model(self, inp_train: list, inp_test: list) -> list:
		"""
		Updates the inner model and returns the predicted test values after fitting on the train
		Args:
			inp_train: input train datasest (a temporal serie)
			inp_test: input test datasest (a temporal serie)

		Returns:
			the predicted test values after fitting on the train ( y_pred_test )
		"""
		self.announce_build()
		self.train_len: int = int(len(inp_train) - self.sample_len)
		self.test_len: int = int(len(inp_test) - self.sample_len)
		return self._build_model(*warp_dataset(inp_train, inp_test))

	def _build_model(self, x_train: list, y_train: list, x_test: list, y_test: list) -> list:
		"""
		Updates the inner model and returns the predicted test values after fitting on the train
		Args:
			x_train: training input data (non-truncated)
			y_train: training correct output data
			x_test: testing input data (non-truncated)
			y_test: testing correct output data

		Returns:
			the predicted test values after fitting on the train ( y_pred_test )
		"""
		x_train = x_train[self.sample_len:self.sample_len + self.train_len]
		x_test = x_test[self.sample_len:self.sample_len + self.test_len]

		self.y_data_train: list = y_train[self.sample_len:self.sample_len + self.train_len]
		self.y_data_test: list = y_test[self.sample_len:self.sample_len + self.test_len]

		self.model = self.model.fit(x_train, self.y_data_train, warmup=0)
		self.y_pred_train: list = self.model.run(x_train)
		self.y_pred_test: list = self.model.run(x_test)
		return self.y_pred_test

	def save_data(self, overwrite: bool = False, *a) -> None:
		"""
		Saves the contents of the qrc to a csv file

		Args:
			overwrite: whether to overwrite the file where we would write data, or just append at the end
			other arguments are ignored
		"""
		dic = {
			'sample_len': self.sample_len,
			'nb_neurons': self.nb_neurons,
			'input_type': self.input_type,
			'train_len': self.train_len,
			'test_len': self.test_len,
			'train_mae': self.get_error_train(),
			'test_mae': self.get_error_test()
		}
		super().save_data(overwrite, dic, f'{constants.MAIN_PATH}/classical_grid.csv', self.header)



#!/usr/bin/env python
"""
Defines reasonable demo-testing functions
"""
from naref.modules.classical.hyper import HyperC
from naref.modules.quantum.hyper import HyperQ
from naref.modules.quantum import QRC
from naref.modules.classical import CRC
from naref.modules import data

def fastest_hs() -> (HyperC, HyperQ):
	"""
	Fastest hyperparameter search
	Returns:
		the completed HyperC and HyperQ object
		all relevant info in csv files
	"""
	hc = HyperC(
		sample_len_list=[8],
		nb_neurons_list=[9, 12, 16, 25, 70],
		input_type_list=['sine'],
		train_len_list=[250, 400],
		verbose=False)
	hc.search()
	hq = HyperQ(
		sample_len_list=[8],
		nb_atoms_list=[4],
		N_samples_list=[1024],
		inp_duration_list=[1000],
		reset_rate_list=[0],
		geometry_list=None,
		atom_distance_list=[15, 20],
		input_type_list=['sine'],
		train_len_list=[250, 400])
	hq.search()
	return hc, hq


def fast_hs() -> (HyperC, HyperQ):
	"""
	Fast hyperparameter search
	Returns:
		the completed HyperC and HyperQ object
		all relevant info in csv files
	"""
	hc = HyperC(
		sample_len_list=[6, 8, 10],
		nb_neurons_list=[9, 12, 16, 25, 70],
		input_type_list=['sine'],
		train_len_list=[250, 400],
		verbose=False)
	hc.search()
	hq = HyperQ(
		sample_len_list=[8],
		nb_atoms_list=[9],
		N_samples_list=[1024],
		inp_duration_list=[1000],
		reset_rate_list=[0],
		geometry_list=None,
		atom_distance_list=[15, 20],
		input_type_list=['sine'],
		train_len_list=[250, 400])
	hq.search()
	return hc, hq


def long_hs() -> (HyperC, HyperQ):
	"""
	Long hyperparameter search
	Returns:
		the completed HyperC and HyperQ object
		all relevant info in csv files
	"""
	hc = HyperC(
		sample_len_list=[8],
		nb_neurons_list=[9, 12, 16, 25, 70],
		input_type_list=['sine', 'mackey'],
		train_len_list=[250, 300, 350, 400, 450, 500, 550])
	hc.search()
	hq = HyperQ()
	hq.search()
	return hc, hq


def demo_notebook():
	"""
	Function used for a typical demonstration.
	It is made to be very fast, the predicted curves are imprecise
	"""
	qrc = QRC(
		sample_len=data.sample_len,
		nb_atoms=4,
		N_samples=data.N_samples,
		inp_duration=data.inp_duration,
		reset_rate=data.reset_rate,
		geometry=data.geometry,
		atom_distance=data.atom_distance,
		input_type=data.input_type,
		test_len=data.real_test_len,
		train_len=data.real_train_len,
		verbose=True)
	qrc.build_model(data.inp_test, data.inp_train)
	qrc.show_test_prediction()
	qrc.show_train_prediction()
	qrc.get_error_train()
	qrc.get_error_test()
	crc = CRC(
		sample_len=data.sample_len,
		nb_neurons=data.nb_neurons)
	crc.build_model(data.inp_test, data.inp_train)
	crc.show_test_prediction()
	crc.show_train_prediction()
	crc.get_error_train()
	crc.get_error_test()

# Example of code :
#
# from naref.demo import *
# from naref.comp import *
# fastest_hs()
# cmp = Comparator(auto_train=False)
# cmp.energy()
# demo_notebook()



#!/usr/bin/env python
"""
Defines the HyperC class, finding hyperparameters for the CRC class
"""
from naref.modules import abstract, data
from naref.modules.classical import CRC


class HyperC(abstract.Hyper):
	def __init__(
			self,
			sample_len_list=None,
			nb_neurons_list=None,
			input_type_list=None,
			train_len_list=None,
			test_len_list=None,
			verbose: bool = True):
		super().__init__(sample_len_list, input_type_list, train_len_list, test_len_list, verbose)
		self.lists.update({
			"nb_neurons": data.nb_neurons_list if nb_neurons_list is None else nb_neurons_list})

	def test_parameters(self, *tup) -> None:
		self.k += 1
		sample_len, input_type, train_len, test_len,  nb_neurons = tup
		real_train_len = train_len - sample_len
		real_test_len = test_len - sample_len

		self.current = CRC(
			nb_neurons=nb_neurons,
			sample_len=sample_len,
			input_type=input_type,
			test_len=real_test_len,
			train_len=real_train_len,
			verbose=self.verbose)
		self.save_parameters()



#!/usr/bin/env python
"""
Defines the HyperQ class, finding hyperparameters for the QRC class
"""
from naref.modules import abstract, data
from naref.modules.quantum import QRC


class HyperQ(abstract.Hyper):
	def __init__(
			self,
			sample_len_list=None,
			nb_atoms_list=None,
			N_samples_list=None,
			inp_duration_list=None,
			reset_rate_list=None,
			geometry_list=None,
			atom_distance_list=None,
			input_type_list=None,
			train_len_list=None,
			test_len_list=None,
			verbose: bool = True):
		super().__init__(sample_len_list, input_type_list, train_len_list, test_len_list, verbose)
		self.lists.update({
			"nb_atoms": data.nb_atoms_list if nb_atoms_list is None else nb_atoms_list,
			"N_samples": data.N_samples_list if N_samples_list is None else N_samples_list,
			"inp_duration": data.inp_duration_list if inp_duration_list is None else inp_duration_list,
			"reset_rate": data.reset_rate_list if reset_rate_list is None else reset_rate_list,
			"geometry": data.geometry_list if geometry_list is None else geometry_list,
			"atom_distance": data.atom_distance_list if atom_distance_list is None else atom_distance_list})

	def test_parameters(self, *tup) -> None:
		self.k += 1
		(
			sample_len, input_type, train_len, test_len, nb_atoms,
			inp_duration, N_samples, reset_rate, geometry, atom_distance
		) = tup
		real_train_len = train_len - sample_len
		real_test_len = test_len - sample_len

		self.current = QRC(
			nb_atoms=nb_atoms,
			geometry=geometry,
			atom_distance=atom_distance,
			N_samples=N_samples,
			sample_len=sample_len,
			reset_rate=reset_rate,
			inp_duration=inp_duration,
			input_type=input_type,
			test_len=real_test_len,
			train_len=real_train_len,
			verbose=self.verbose)
		self.save_parameters()



#!/usr/bin/env python
"""
Defines various global variables
"""

# Default verbosity level for all Forecaster objects
VERBOSE = True

# Working directory for datasets
MAIN_PATH = "./results"

# Interval between prints when training
FEEDBACK_INTERVAL = 10

# Default random seed (for numpy, random, and reservoirpy naref.modules)
DEFAULT_SEED = 0

# Other examples : MockDevice, IroiseMVP, AnalogDevice
from pulser.devices import Chadoq2
PULSER_DEVICE = Chadoq2

# Other examples : mean_squared_log_error, ...
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
ERROR_METRIC = mean_absolute_error
try:
	ERROR_NAME = ERROR_METRIC.__doc__.split('\n')[0][:-1]
except:
	ERROR_NAME = "Error"
# All errors considered for benchmarking.
ERR_BENCHMARK = {
	'mae': mean_absolute_error,
	'rmse': (lambda x, y: mean_squared_error(x, y) ** .5),
	'r2': r2_score}



#!/usr/bin/env python
"""
Defines the QRC class and its abstract classes
"""

from naref.modules import constants, topology, abstract
import numpy as np
from numpy import random as rd
import pandas as pd
from pulser import Pulse, Sequence, Register
from pulser_simulation import QutipEmulator
import time
import warnings

rd.seed(constants.DEFAULT_SEED)
warnings.filterwarnings('ignore')


class QRCTrainer(abstract.Forecaster):

	def __init__(
			self,
			sample_len: int,
			nb_atoms: int,
			N_samples: int,
			inp_duration: int,
			reset_rate: float,
			geometry: str,
			atom_distance: float,
			input_type: str = 'unknown',
			train_len: int = -1,
			test_len: int = -1,
			verbose: bool = constants.VERBOSE):
		super().__init__(input_type, train_len, test_len, verbose)
		self.name = "QRC"
		self.longname = "Quantum Reservoir"
		self.inp_duration: int = inp_duration  # the reservoir will see each input for this pulse duration (in ns)
		self.N_samples: int = N_samples  # number of shots
		self.sample_len: int = sample_len  # length of input time series
		self.reset_rate: float = reset_rate  # probability of resetting the whole reservoir (for ESP)
		self.train_data: list = []  # contains the latest provided training data
		self.rese_data: list = []  # contains the training data of the reservoir
		self.wlist: list = []  # list of input training data (truncated to fit a window)
		self.clist: list = []  # counts (of sample of window) list
		self.nclist: list = []  # normalised counts (of sample of window) list
		self.ncdf: pd.DataFrame = None  # normalised counts Panda DataFrame, filled with 0 where was NaN
		self.X_all_train: np.ndarray = None  # array version of the above. We compute w by solving w = X^-1 * data.
		self.w: np.ndarray = None  # Reservoir output weights.

		self.n: int = nb_atoms
		self.geom: str = geometry
		self.atom_distance: float = atom_distance
		self.seq: Sequence = None

		self.__k_time: float = 0  # Time since the last print of training time
		self.scope: str = "training"  # Current action

	def __reset_seq(self) -> None:
		""" Sets self.set to an initial, blank sequence, with the right layout, device and channels """
		self.seq = Sequence(self.layout(), constants.PULSER_DEVICE)
		self.seq.declare_channel("global", "rydberg_global")

	def layout(self) -> Register:
		""" Returns the right Register object associated with the lattice of geometry 'self.geom' """

		if self.geom == 'grid_lattice':
			q_dict = topology.grid_lattice(int(np.sqrt(self.n)), int(np.sqrt(self.n)), self.atom_distance)
		elif self.geom == 'grid_lattice_centred':
			q_dict = topology.grid_lattice_centred(int(np.sqrt(self.n)), self.atom_distance)
		elif self.geom == 'circle_lattice':
			q_dict = topology.circle_lattice(self.n, self.atom_distance)
		elif self.geom == 'triangle_lattice':
			q_dict = topology.triangle_lattice(int(-1 / 2 + np.sqrt(1 / 4 + 2 * self.n)), self.atom_distance)
		else:
			raise Exception("Unknown Lattice Type")

		reg = Register(q_dict)
		return reg

	def build_reservoir(self, inp_trunc: list, scope: str = "training") -> list:
		"""
		Encodes input_trunc as input parameters in local pulses and constructs the reservoir from the encoding.
		Returns all measurement outcomes associated with the truncated input

		The input is here called truncated, as typically, the total dataset is divided into training and test set.
		Each can therefore be considered as 'truncated'.

		Args:
			inp_trunc: truncated input (cf above)
			scope: whether to print that we are in the training or testing phase

		Returns:
			the list [lists of the samples of Z_MEASUREMENT( R(window) ) (for each window)]
		"""

		self.rese_data = inp_trunc
		self.scope = scope

		# total time since the last constants.FEEDBACK_INTERVAL steps
		self.__k_time = 0

		# number of available windows
		nwin = len(self.rese_data) - self.sample_len

		# list of outcomes for each window (uK ...uk+tau)
		return [self.window(k) for k in range(nwin)]

	def window(self, k: int) -> dict:
		"""
		For each window (uK ...uK+tau) during the execution,
		we have to do the sequence of pulses corresponding to R(windowK, rhoK)

		Args:
			k: index of the window (uK ...uK+tau)

		Returns:
			the dict of the samples of Z_MEASUREMENT( R(window) )
		"""

		if self.verbose:
			tic = time.time()  # measure the time spent in this window

		# initialize Sequence self.seq
		self.__reset_seq()

		# extract subsection of time series to pass into the reservoir
		self.wlist = self.rese_data[k:k + self.sample_len]

		# for each ui in (uK ...uK+tau)
		for ui in range(self.sample_len):
			self.add_to_sequence(ui)

		sim = QutipEmulator.from_sequence(self.seq, sampling_rate=1)
		results = sim.run(progress_bar=False)  # run the simulation
		outcomes = results.sample_final_state(N_samples=self.N_samples)

		if self.verbose:
			toc = time.time()
			self.__k_time += (toc - tic)
			if k % 10 == 9:  # as k starts at 0, this ensures 10 steps have passed before the first feedback
				print(f"{k + 1} {self.scope} steps finished, {self.scope} time : {self.__k_time:.3f}s")
				self.__k_time = 0

		return outcomes

	def add_to_sequence(self, i: int) -> None:
		"""
		For each ui in a given window (uK ...uK+tau), we have to add either a corresponding pulse or a reset.
		Args:
			i: index of the input under scrutiny
		"""
		duration = self.inp_duration  # duration for which the reservoir will see the input = pulse duration (in ns)

		reset = rd.choice(
			np.arange(0, 2),
			p=[1 - self.reset_rate, self.reset_rate],
			size=1)  # in case we want to add random noise

		if reset:
			print("reset")
			self.__reset_seq()

		else:
			amplitude = self.wlist[i]  # input is encoded in the pulse amplitude
			detuning = 0

			# It seems that Pulser sometimes messes up the random seed at this point.
			if np.isclose(self.reset_rate, 0):
				# If we do not need randomness at this point, we reinitialise the seed.
				rd.seed(constants.DEFAULT_SEED)

			pulse = Pulse.ConstantPulse(duration, amplitude, detuning, phase=0.00, post_phase_shift=0.0)
			self.seq.add(pulse, "global")

	def train(self, data: list, train_len: int) -> float:
		"""
		Executes the training for all inputs in data.
		Fills in self.train_data.
		Constructs the optimised weights from the training.
		Returns and displays the final total error (train error).

		Args:
			data: input data
			train_len: the number of training steps
		"""
		self.train_data = data[:self.sample_len + train_len]

		total_time = 0
		tic = time.time()
		self.clist = self.build_reservoir(self.train_data)

		# normalise the counts list so that it is coherent between train and test
		self._fill_nclist(self.clist)

		# The following link justifies using from_dict for a list of dict objects
		# www.geeksforgeeks.org/create-a-pandas-dataframe-from-list-of-dicts/
		self.ncdf = pd.DataFrame.from_dict(self.nclist).fillna(0)
		self.X_all_train = self.ncdf.to_numpy()

		# calculate pseudo inverse
		xp = np.linalg.pinv(self.X_all_train)
		# calculate the weight vector (linear regression)
		self.w = np.dot(xp, self.train_data[self.sample_len:self.sample_len + len(self.clist)])

		toc = time.time()
		total_time += (toc - tic)
		print(f"Training finished. Total training time: {total_time:.3f}s")

		self.y_data_train = self.train_data[self.sample_len:self.sample_len + train_len]
		self.y_pred_train = self._get_prediction(self.X_all_train, len(self.clist))
		error_train: float = self.get_error_train()
		print(f"Train error ({constants.ERROR_NAME}): ", error_train)
		return error_train

	def _get_prediction(self, X_all: np.ndarray, c_len: int) -> list:
		"""
		Args:
			X_all: array version of normalized counts (after pd filling the NaN)
			c_len: length of the list of counts

		Returns:
			List of the predictions associated with the reservoir weights w
		"""
		return [np.dot(X_all[c, :], self.w) for c in range(c_len)]

	def _fill_nclist(self, counts: list) -> None:
		"""
		Normalizes the counts list so that it is coherent between the testing and training phases.
		Args:
			counts: list of counts (of sample of window), either from training or testing phase.
		"""
		self.nclist = [{k: v / self.N_samples for k, v in c.items()} for c in counts]


class QRCTester(QRCTrainer):

	def __init__(
			self,
			sample_len: int,
			nb_atoms: int,
			N_samples: int,
			inp_duration: int,
			reset_rate: float,
			geometry: str,
			atom_distance: float,
			input_type: str = 'unknown',
			train_len: int = -1,
			test_len: int = -1,
			verbose: bool = constants.VERBOSE):
		super().__init__(
			sample_len, nb_atoms, N_samples, inp_duration, reset_rate, geometry, atom_distance,
			input_type, train_len, test_len, verbose)
		self.test_data: list = []  # contains the latest provided test data
		self.clist_test: list = []  # list of counts (of sample of window) from the testing phase
		self.X_all_test: np.ndarray = None  # array computing ...

	def evaluate(self, test_data: list, test_len: int) -> float:
		"""
		Builds the reservoirs associated with the test data. Returns the observed hidden states and the predictions
		Args:
			test_data: input data
			test_len: the number of testing steps
		"""

		self.test_data = test_data[:self.sample_len + test_len]
		self.clist_test = self.build_reservoir(self.test_data, scope="testing")

		# normalise the counts list so that it is coherent between train and test
		self._fill_nclist(self.clist_test)

		df_copy = self.ncdf.iloc[:0, :].copy().fillna(0)
		df_copy = df_copy.append(self.nclist, ignore_index=True, sort=False).fillna(0)
		self.X_all_test = df_copy.to_numpy()

		# evaluate testing error
		self.y_pred_test = self._get_prediction(self.X_all_test, len(self.clist_test))
		self.y_data_test = self.test_data[self.sample_len:self.sample_len + test_len]

		error_test: float = self.get_error_test()
		print(f"Test error ({constants.ERROR_NAME}): ", error_test)
		return error_test


class QRC(QRCTester):
	header = [
		'input_type', 'train_len', 'test_len', 'inp_duration', 'N_samples', 'sample_len',
		'reset_rate', 'nb_atoms', 'geometry', 'atom_distance', 'train_mae', 'test_mae']

	def __init__(
			self,
			sample_len: int,
			nb_atoms: int,
			N_samples: int,
			inp_duration: int,
			reset_rate: float,
			geometry: str,
			atom_distance: float,
			input_type: str = 'unknown',
			train_len: int = -1,
			test_len: int = -1,
			verbose: bool = constants.VERBOSE):
		super().__init__(
			sample_len, nb_atoms, N_samples, inp_duration, reset_rate, geometry, atom_distance,
			input_type, train_len, test_len, verbose)

	def build_model(self, inp_train: list, inp_test: list) -> list:
		"""
		Updates the inner model and returns the predicted test values after fitting on the train
		Args:
			inp_train: input train datasest (a temporal serie)
			inp_test: input test datasest (a temporal serie)

		Returns:
			the predicted test values after fitting on the train ( y_pred_test )
		"""
		self.announce_build()
		self.train(inp_train, train_len=self.train_len)
		self.evaluate(inp_test, test_len=self.test_len)
		return self.y_pred_test

	def save_data(self, overwrite: bool = False, *a) -> None:
		"""
		Saves the contents of the qrc to a csv file

		Args:
			overwrite: whether to overwrite the file where we would write data, or just append at the end
			other arguments are ignored
		"""
		dic = {
			'inp_duration': self.inp_duration,
			'N_samples': self.N_samples,
			'sample_len': self.sample_len,
			'reset_rate': self.reset_rate,
			'nb_atoms': self.n,
			'geometry': self.geom,
			'atom_distance': self.atom_distance,
			'input_type': self.input_type,
			'train_len': self.train_len,
			'test_len': self.test_len,
			'train_mae': self.get_error_train(),
			'test_mae': self.get_error_test()}
		fname = f'{constants.MAIN_PATH}/{self.n}_{self.input_type}_{self.geom}.csv'
		super().save_data(overwrite, dic, fname, self.header)



#!/usr/bin/env python
"""
Defines the auxiliary functions handling comparison methods
"""
from naref.modules import constants
from matplotlib import pyplot as plt


def errors(vals: iter, preds: iter):
	"""
	Typically, ERR_BENCHMARK contains 'mae' 'rmse' and 'r2' errors.
	Args:
		vals: an iterable of floats
		preds: an iterable of floats

	Returns:
		a dictionary containing errors obtained when comparing vals & preds (for all errors in ERR_BENCHMARK)
	"""

	return {k: v(vals, preds) for k,v in constants.ERR_BENCHMARK.items()}


# Emission costs (tCO2 eq/hour)
C_Fresnel = (2.1*3.6)/(1000*2.1)
C_Rubi = (4.2*3)/(1000*0.4)
C_GPU = 0.09/1000
C_Joliot = 215/1000
# Required computation time as specified in the given Excel document
T_QPU, T_GPU, T_Joliot = 458, 884000, 67
# Classical processing unit frequencies
f_Joliot, f_GPU = 7 * 1e15, 5 * 1e11


def print_former_benchmarks():
	"""
	From PASQAL data, we noted above the emissions of each type of device in tCO2 eq/hour.
	We can print that it is coherent with the given Excel document
	"""
	print("Emissions from Fresnel = {} eq tCO2".format(T_QPU*C_Fresnel))
	print("Emissions from Rubi = {} eq tCO2".format(T_QPU*C_Rubi))
	print("Emissions from GPU cluster = {} eq tCO2".format(T_GPU*C_GPU))
	print("Emissions from Joliot Curie = {} eq tCO2".format(T_Joliot*C_Joliot))


# useful functions for energy metrics
def fct_L(L_tilde):
	return L_tilde ** (1 / 3)


def fct_S(L):
	return fct_L(L) ** 0.5


def show_prediction(pred: list, real: list, phase: str, longname: str) -> None:
	"""
	Shows how close are prea and real, plottable data
	"""
	plt.figure(figsize=(10, 3))
	plt.title(f"{longname} {phase} Prediction")
	plt.xlabel("$t$")
	plt.plot(pred, label="Predicted", color="blue")
	plt.plot(real, label="Real", color="red")
	plt.legend()
	plt.show()



#!/usr/bin/env python
# Copyright 2023 NAREF Development Team
#
# Unless required by applicable law or agreed to in writing, this software
# is on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied.

"""A quantum & classical reservoir handler for Temporal Series Prediction."""

from naref.modules.comp import Comparator
from naref.modules.quantum.hyper import HyperQ
from naref.modules.classical.hyper import HyperC
from naref.modules.demo import fastest_hs
from naref.modules import *
